#include "Decompress.h"

static inline u32 GetNextBit(u8 &currentBits, u8 &bitCnt, const u8 * source, u32 &curLen)
{
	if (bitCnt)
	{
		// If there are bits left then shift up by one
		currentBits = currentBits << 1;
		bitCnt--;
	}
	else
	{
		currentBits = (u32)source[curLen];
		curLen++;
		bitCnt=7;
	}

	return (u32) ((currentBits >> 7) & 1);
}

int Decompress( const u8 * source, u32  sourceLen,u8 * dest, u32 * destLen )
{
	u8 currentBits = 0;
	u32 curLen = 0, outLen = 0, lastworkOffset = 1;
	u8 bitCnt = 0;

	// First bit (source[0] & 0x80) encoded should always be 1 to indicate a literal byte because obviously we have no previous literals to copy from since we have no uncompressed data.
	// Future expansion could encode a 0 bit and use that as a flag for something.

	for (;;)
	{
		while (GetNextBit(currentBits,bitCnt,source,curLen))
		{
			dest[outLen] = source[curLen];
			outLen++;
			curLen++;
		}

		u32 workOffset = 1;

		do
		{
			workOffset = (workOffset<<1) + GetNextBit(currentBits,bitCnt,source,curLen);
		} while (!GetNextBit(currentBits,bitCnt,source,curLen));

		if (workOffset == 2)
		{
			workOffset = lastworkOffset;
		}
		else
		{
			workOffset = ((workOffset-3)<<8) + source[curLen];
			curLen++;
			if (workOffset == 0xffffffffU)
			{
				break;
			}
			workOffset++;
			lastworkOffset = workOffset;
		}

		u32 workLen = GetNextBit(currentBits,bitCnt,source,curLen);
		workLen = (workLen<<1) + GetNextBit(currentBits,bitCnt,source,curLen);

		if (workLen == 0)
		{
			workLen++;
			do
			{
				workLen = (workLen<<1) + GetNextBit(currentBits,bitCnt,source,curLen);
			} while (!GetNextBit(currentBits,bitCnt,source,curLen));
			workLen += 2;
		}

		if (workOffset > LONG_OFFSET_THRESHOLD)
		{
			workLen++;
		}

		const u8 *workPos = dest + outLen - workOffset;
		dest[outLen] = *workPos;
		outLen++;
		workPos++;
		do
		{
			dest[outLen] = *workPos;
			outLen++;
			workPos++;
			workLen--;
		} while (workLen > 0);
	}
	*destLen = outLen;
	return curLen == sourceLen ? (GBA_E_OK) : (GBA_E_ERROR);
}
